# Copyright 2017 John Reese
# Licensed under the MIT license

import os.path
import logging

from pathlib import Path
from importlib import import_module
from types import ModuleType
from typing import List

log = logging.getLogger(__name__)


def import_units() -> List[ModuleType]:
    """Find and import units in this path."""
    modules: List[ModuleType] = []
    root = Path(Path(__file__).parent)  # appease mypy, Path.parents -> PurePath
    log.debug(f"Searching for units in {root}...")
    for path in root.glob("*.py"):
        name = path.stem
        log.debug(f"Loading unit {name}")
        module = import_module(f"edi.units.{name}")
        modules.append(module)
    return modules



# Copyright 2017 John Reese
# Licensed under the MIT license

import logging

from aioslack import Event
from datetime import datetime
from pathlib import Path

from ..core import Unit

log = logging.getLogger(__name__)


class ChatLog(Unit):

    async def start(self) -> None:
        self.root = Path.home() / "slack" / "logs"
        self.root.mkdir(parents=True, exist_ok=True)

    def log_message(self, channel: str, dt: datetime, message: str) -> None:
        date = dt.strftime(r"%Y-%m-%d")
        time = dt.strftime(r"%H:%M:%S")

        log.info(f"[{date} {time}] #{channel} {message}")

        filename = self.root / channel / f"{date}.log"
        filename.parent.mkdir(parents=True, exist_ok=True)
        with open(filename, "a") as f:
            f.write(f"[{time}] {message}\n")

    async def on_hello(self, event: Event) -> None:
        assert event
        self.root /= self.slack.team.name
        log.info(f"logging messages to {self.root}")

    async def on_message(self, event: Event) -> None:
        dt = datetime.fromtimestamp(float(event.ts))

        channel = self.slack.channels[event.channel].name
        username = self.slack.users[event.user].name

        if "subtype" in event:
            subtype = event.subtype

            if subtype == "me_message":
                message = f"* {username} {event.text}"
            elif subtype == "channel_join":
                message = f"* {username} joined the channel"
            elif subtype == "channel_leave":
                message = f"* {username} left the channel"
            elif subtype == "channel_purpose":
                message = f"* {username} set the channel purpose: {event.purpose}"
            elif subtype == "channel_topic":
                message = f"* {username} set the channel topic: {event.topic}"
            elif subtype == "channel_name":
                message = (
                    f"* {username} renamed the channel from "
                    f"{event.old_name} to {event.name}"
                )

        else:
            message = f"<{username}> {event.text}"

        self.log_message(channel, dt, message)

    async def on_reaction_added(self, event: Event) -> None:
        ts = float(event.item["ts"])
        dt = datetime.fromtimestamp(ts)
        channel = event.item["channel"]
        history = await self.slack.api(
            "channels.history", channel=channel, latest=ts, oldest=ts, inclusive=True
        )
        context = Event.generate(history.messages[0])
        channel = self.slack.channels[channel].name
        reactor = self.slack.users[event.user].name
        username = self.slack.users[context.user].name
        text = context.text
        if len(text) > 40:
            text = text[:40].rsplit(" ", 1)[0] + "..."
        message = f" * {reactor} reacted :{event.reaction}: to <{username}> {text}"

        self.log_message(channel, dt, message)



# Copyright 2017 John Reese
# Licensed under the MIT license

import asyncio
import logging
import signal

try:
    import uvloop
except ImportError:
    uvloop = None

from aioslack import Event, Slack
from ent import Singleton
from typing import List

from .config import Config
from .core import Unit
from .log import init_logger
from .units import import_units

log = logging.getLogger(__name__)


class Edi(metaclass=Singleton):
    """Main event framework for the bot."""

    loop: asyncio.AbstractEventLoop

    def __init__(self, config: Config) -> None:
        log.debug(f"initializing with {config}")
        self.config = config
        self.units: List[Unit] = []

        self._started = False

    def start(self) -> None:
        """Start the asyncio event loop, and close it when we're done."""
        if self._started:
            raise Exception("Edi already started")

        if uvloop and self.config.uvloop:
            log.debug("Using uvloop")
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

        self.loop = asyncio.new_event_loop()

        if self.config.debug:
            self.loop.set_debug(True)

        self.loop.add_signal_handler(signal.SIGINT, self.sigterm)
        self.loop.add_signal_handler(signal.SIGTERM, self.sigterm)

        asyncio.ensure_future(self.run(), loop=self.loop)
        self.loop.run_forever()
        self.loop.close()

    def sigterm(self) -> None:
        """Handle Ctrl-C or SIGTERM by stopping the event loop nicely."""
        log.warning("Signal received, stopping execution")
        asyncio.ensure_future(self.stop(), loop=self.loop)

    async def run(self) -> None:
        """Execute all the bits of Edi."""

        try:
            self.slack = Slack(token=self.config.token)

            import_units()
            self.units = [unit(self.slack) for unit in Unit.all_units()]
            log.debug(f"Starting {len(self.units)} units")
            await asyncio.gather(*[unit.start() for unit in self.units])

            log.debug(f"Starting RTM loop")
            async for event in self.slack.rtm():
                await self.dispatch(event)

        except Exception:
            log.exception(f"run loop exception, stopping")
            await self.stop()
            raise

    async def stop(self) -> None:
        """Stop all the bits of Edi."""

        try:
            log.debug(f"Stopping {len(self.units)} units")
            await asyncio.gather(
                *[unit.stop() for unit in self.units], return_exceptions=True
            )

            await self.slack.close()

        finally:
            self.loop.stop()
            log.info("Goodbye!")

    async def dispatch(self, event: Event) -> None:
        """Dispatch events to all active units."""
        results = await asyncio.gather(
            *[unit.dispatch(event) for unit in self.units], return_exceptions=True
        )

        for result in results:
            if isinstance(result, BaseException):
                log.error(f"uncaught exception:\n{result}")


def init_from_config(config: Config) -> None:
    """Initialize Edi from a loaded `Config` object."""

    init_logger(stdout=True, file_path=config.log, debug=config.debug)
    log.debug("logger initialized")

    Edi(config).start()



# Copyright 2017 John Reese
# Licensed under the MIT license

import logging
import sys


def init_logger(
    stdout: bool = True, file_path: str = None, debug: bool = False
) -> logging.Logger:
    """Initialize the logging system for stdout and an optional log file."""

    log = logging.getLogger("")

    level = logging.DEBUG if debug else logging.INFO
    log.setLevel(level)

    logging.addLevelName(logging.ERROR, "E")
    logging.addLevelName(logging.WARNING, "W")
    logging.addLevelName(logging.INFO, "I")
    logging.addLevelName(logging.DEBUG, "V")

    date_fmt = r"%H:%M:%S"
    stdout_fmt = "%(levelname)s: %(message)s"
    verbose_fmt = (
        "%(asctime)s,%(msecs)d %(levelname)s "
        "%(module)s:%(funcName)s():%(lineno)d   "
        "%(message)s"
    )

    handler: logging.Handler

    if stdout:
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(level)

        if debug:
            handler.setFormatter(logging.Formatter(verbose_fmt, date_fmt))
        else:
            handler.setFormatter(logging.Formatter(stdout_fmt, date_fmt))

        log.addHandler(handler)

    if file_path:
        handler = logging.FileHandler(file_path)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(logging.Formatter(verbose_fmt, date_fmt))

        log.addHandler(handler)

    return log
